/* slurp-download.h
 *
 * Copyright (C) 2017 Julian Richen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define G_LOG_DOMAIN "slurp-download.h"

#include "slurp-download.h"

/**
 * SECTION:slurp-download
 * @title: SlurpDownload
 * @short_description: Asynchronous resource download using the libsoup package
 * @include: slurp-download.h
 * @see_also: #SoupSession, #SoupRequest, #GMainLoop, #GMainContext,
 *            #GCallback, #GCancellable, #GFile, #GInputStream
 *
 * Downloads files in parallel by running asynchronous file download with the
 * help of libsoup. Each download runs it's own #GMainLoop but refs the current
 * #GMainContext, if one exists using g_main_context_ref_thread_default(). If no
 * current #GMainContext exists then a new one is created. If you plan on using
 * downloads inside of an application you should run that application first
 * either with a #GMainLoop or using #GApplication so I can hook onto the
 * main context.
 **/

typedef struct _DownloadResourceData {
    GMainLoop *loop;

    gchar *uri;
    gchar *path;
    gboolean overwrite;

    GCancellable *cancellable;
    DownloadResourceDataProgress p_handler;
    gpointer p_user_data;
    DownloadResourceDataCallback c_handler;
    gpointer c_user_data;

    GFile *file;
    gchar buffer[16384]; // 16 * 1024
    GOutputStream *output;

    guint64 total_bytes;
    guint64 downloaded_bytes;
    guint64 last_progress_time;
} DownloadResourceData;

struct _SlurpDownload
{
    GObject parent_instance;
};

typedef struct
{
    gchar *default_directory;
    gboolean default_overwrite;
} SlurpDownloadPrivate;

G_DEFINE_TYPE_WITH_PRIVATE (SlurpDownload, slurp_download, G_TYPE_OBJECT)

enum {
    PROP_0,
    PROP_DEFAULT_DIRECTORY,
    PROP_DEFAULT_OVERWRITE,
    N_PROPS
};

static GParamSpec *properties[N_PROPS];

enum {
    ADDED,
    REMOVED,
    FINISHED,
    CANCELLED,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL];

static void
slurp_download_finalize (GObject *object)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    if (priv->default_directory)
        g_free (priv->default_directory);

    G_OBJECT_CLASS (slurp_download_parent_class)->finalize (object);
}

static void
slurp_download_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_DEFAULT_DIRECTORY:
            g_value_set_string (value, priv->default_directory);
            break;

        case PROP_DEFAULT_OVERWRITE:
            g_value_set_boolean (value, priv->default_overwrite);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_download_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_DEFAULT_DIRECTORY:
            priv->default_directory = g_value_dup_string (value);
            break;

        case PROP_DEFAULT_OVERWRITE:
            priv->default_overwrite = g_value_get_boolean (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_download_class_init (SlurpDownloadClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = slurp_download_finalize;
    object_class->get_property = slurp_download_get_property;
    object_class->set_property = slurp_download_set_property;

	/**
	 * SlurpDownload:directory:
	 *
	 * The default directory to save resources into.
     *
     * Default: $XDG_USER_DIR/DOWNLOAD (ex. $HOME/Downloads)
	 */
    properties[PROP_DEFAULT_DIRECTORY] =
        g_param_spec_string ("directory",
                             "Default directory",
                             "The default directory to save resources into.",
                             g_get_user_special_dir (G_USER_DIRECTORY_DOWNLOAD),
                             G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT_ONLY);

	/**
	 * SlurpDownload:overwrite:
	 *
	 * Should files be overwritten by default.
     *
     * Recommended: %FALSE
	 */
    properties[PROP_DEFAULT_OVERWRITE] =
        g_param_spec_boolean ("overwrite",
                              "Overwrite files",
                              "Should files be overwritten by default? Default: FALSE",
                              FALSE,
                              G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT_ONLY);

    g_object_class_install_properties (object_class,
                                       N_PROPS,
                                       properties);

    signals[ADDED] =
        g_signal_new ("added",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_STRING);

    signals[REMOVED] =
        g_signal_new ("removed",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_STRING);

    signals[FINISHED] =
        g_signal_new ("finished",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_STRING);

    signals[CANCELLED] =
        g_signal_new ("cancelled",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_STRING);
}

static void
slurp_download_init (SlurpDownload *self)
{
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
SlurpDownload *
slurp_download_new (void)
{
    return g_object_new (SLURP_TYPE_DOWNLOAD, NULL);
}

void
slurp_download_add_uri_full (SlurpDownload                *self,
                             gchar                        *uri,
                             gchar                        *path,
                             gboolean                      overwrite,
                             GCancellable                 *cancellable,
                             DownloadResourceDataProgress  p_handler,
                             gpointer                      p_user_data,
                             DownloadResourceDataCallback  c_handler,
                             gpointer                      c_user_data)
{

}

void
slurp_download_add_uri_with_path (SlurpDownload *self,
                                  gchar         *uri,
                                  gchar         *path,
                                  GCancellable  *cancellable)
{

}

void
slurp_download_add_uri (SlurpDownload *self,
                        gchar         *uri,
                        GCancellable  *cancellable)
{
    g_print ("Added file: \"%s\"\n", uri);
}
